import os
import traceback
import numpy as np
import pandas as pd
import tkinter.filedialog
from sys import executable, exit

# currentDirectory = os.path.dirname(executable)
currentDirectory = os.path.dirname(__file__)

def PrintLogo ():
    try:
        logoPath = os.path.join(currentDirectory, "Resources", "logo.txt")
        with open(logoPath, "r") as logo:
            print(logo.read() + "\n")
    except: print("Could not read " + logoPath)

def Exit (message="", error=None, immediate=False):
    if error: message = "ERROR OCCURRED\n" + message + "\n" + traceback.format_exc()
    if message: print(message)
    if not immediate: input("Press enter to continue . . .")
    exit(1)

class Load ():
    def __init__ (self, defaultDirectory):
        self.grades = pd.DataFrame()
        self.courses = []
        self.defaultDirectory = defaultDirectory
        self.gradePath = ""
        self.Load()

    def IsCorrectYearDirectory (self, directory): return len(directory) == 9 and directory[:4].isnumeric() and directory[4] == "-" and directory[5:].isnumeric()
    def IsCorrectQuarterDirectory (self, directory): directory = directory.split(" "); return len(directory) == 2 and directory[0] == "Quarter" and directory[1].isnumeric();

    def GetLatestGradeCSVs (self):
        #* Find latest year directory
        cumulativePath = currentDirectory + os.sep + "Grades" + os.sep
        yearDirectories = [directory for directory in os.listdir(cumulativePath) if os.path.isdir(cumulativePath + directory) and self.IsCorrectYearDirectory(directory)]
        if len(yearDirectories) == 0: print("No year directories were found."); return [];
        yearDirectories.sort(key=lambda directory: int(directory.split("-")[1]))
        cumulativePath += yearDirectories[-1] + os.sep
        #* Find latest quarter directory
        quarterDirectories = [directory for directory in os.listdir(cumulativePath) if os.path.isdir(cumulativePath + directory) and self.IsCorrectQuarterDirectory(directory)]
        if len(quarterDirectories) == 0: print("No quarter directories were found."); return [];
        quarterDirectories.sort(key=lambda directory: int(directory.split(" ")[1]))
        cumulativePath += quarterDirectories[-1] + os.sep
        #* Find gradeCSVs
        gradeCSVs = [cumulativePath + file for file in os.listdir(cumulativePath) if os.path.isfile(cumulativePath + file) and file.endswith(".csv")]
        if len(gradeCSVs) == 0: print("No Grade CSVs were found."); return [];
        return gradeCSVs
    
    def FindDefaultDirectory (self):
        if self.defaultDirectory.lower().strip() == "latest":
            gradeCSVs = self.GetLatestGradeCSVs()
            match len(gradeCSVs):
                case 0: self.defaultDirectory = currentDirectory + os.sep + "Grades" #? Use Grades if no gradeCSVs
                case 1: self.gradePath = gradeCSVs[0] #? If there is only one gradeCSV, use that.
                case _: self.defaultDirectory = os.path.dirname(gradeCSVs[0]) #? Directory containing the gradeCSVs
            return len(gradeCSVs)

        #* Specified defaultDirectory
        if not os.path.isabs(self.defaultDirectory):
            self.defaultDirectory = currentDirectory + os.sep + "Grades" + os.sep + self.defaultDirectory
        if not os.path.exists(self.defaultDirectory):
            print("defaultDirectory, " + self.defaultDirectory + ", does not exist.")
            self.defaultDirectory = currentDirectory + os.sep + "Grades"
        return -1

    def SelectGradeCSV (self):
        if self.FindDefaultDirectory() == 1: return #? Opening default gradeCSV
        #* Create File Dialog
        root = tkinter.Tk()
        root.withdraw()
        root.attributes("-topmost", True)
        try: self.gradePath = tkinter.filedialog.askopenfile(title="Select Grade CSV", defaultextension=".csv", filetypes=[("CSV files", "*.csv"), ("All files", "*.*")], initialdir=self.defaultDirectory).name.replace("/", os.sep)
        except AttributeError as e: Exit("No file selected.", immediate=True)
        except Exception as e: Exit("An error occurred while selection the Grade CSV.", e)
        root.destroy()

    def GetCSVInfo (self):
        if "Grades" + os.sep in self.gradePath: path = "".join(self.gradePath.split("Grades" + os.sep)[-1]).split(os.sep)
        else: print("Could not find the Grades directory in this file path: " + self.gradePath + "."); return "";
        if len(path) != 3 or not self.IsCorrectYearDirectory(path[0]) or not self.IsCorrectQuarterDirectory(path[1]): print("Could not read the file path: " + self.gradePath + "."); return "";
        return "\n" + path[0] + " " + path[1] + " " + (".".join(path[2].split(".")[:-1]) if "." in path[2] else path[2])

    def ReadGradeCSV (self):
        try: self.grades = pd.read_csv(self.gradePath)
        except FileNotFoundError as e: Exit(self.gradePath + " could not be found.", e)
        except pd.errors.ParserError as e: Exit("It seems the " + self.gradePath + " file is not formatted correctly. Fix the file and try running this program again.", e)
        except Exception as e: Exit("An error has occurred while trying to read " + self.gradePath + ".", e)

    def ConvertDateColumn (self):
        try:
            self.grades["Date"] = pd.to_datetime(self.grades["Date"], format="%b %d %Y")
            #* Remove duplicate dates
            uniqueDates = []
            for index, date in enumerate(self.grades["Date"]):
                if date in uniqueDates: self.grades.drop(index, inplace=True)
                else: uniqueDates.append(date)
            self.grades.sort_values("Date", inplace=True)
            self.grades.reset_index(drop=True, inplace=True)
        except KeyError as e: Exit("There was no \"Date\" column in " + self.gradePath + ".", e)
        except ValueError as e: Exit("Failed to read date.", e)
        except Exception as e: Exit("There was an error while trying to read the Date column", e)

    def GetGradedCourses (self):
        try:
            self.courses = [course for course in list(self.grades.columns)[1:] if not all(np.isnan(grade) for grade in self.grades[course])] #? Filtering courses that have no grades yet
            #* Propagation for empty grades
            for course in self.courses:
                previousGrade = -np.inf
                for index in range(len(self.grades[course]) - 1, -1, -1): #? Loop backwards through grades in each course
                    if np.isnan(self.grades[course][index]): self.grades.loc[:, (course, index)] = previousGrade #? Fill in empty past grades with closest future grade
                    elif previousGrade == -np.inf and index != len(self.grades[course]) - 1: self.grades[course][index+1:] = [self.grades[course][index]] * (len(self.grades[course]) - index - 1) #? Fill in future grades with most recent past grade
                    previousGrade = self.grades[course][index]
            self.courses = [course for course in self.courses if len(self.grades[course]) >= 2] #? Filter out courses that would be nothing or a single point
            if len(self.courses) == 0: Exit("There are not enough grades yet for any courses to be displayed. Come back in a few days when some data has been recorded.")
        except IndexError as e: Exit("No courses could be found.", e)
        except Exception as e: Exit("There was an error while trying to get the courses.", e)

    def Load (self):
        self.SelectGradeCSV()
        self.ReadGradeCSV()
        self.ConvertDateColumn()
        self.GetGradedCourses()